Object subclass: #SoundRecorder
	instanceVariableNames: 'stereo samplingRate recordLevel recordedBuffers recordedSound recordProcess bufferAvailableSema paused meteringBuffer meterLevel meterRead soundPlaying currentBuffer nextIndex codec audioFile '
	classVariableNames: 'CanRecordWhilePlaying '
	poolDictionaries: ''
	category: 'Sound-Synthesis'!

!SoundRecorder methodsFor: 'initialization' stamp: 'jm 11/19/2003 12:42'!
initialize
	"SoundRecorder new"

	stereo _ false.
	samplingRate _ 22050.
	recordLevel _ 0.5.
	self initializeRecordingState.
! !

!SoundRecorder methodsFor: 'initialization' stamp: 'jm 5/12/2006 09:39'!
initializeRecordingState

	recordProcess _ nil.
	bufferAvailableSema _ nil.
	paused _ true.
	meteringBuffer _ nil.
	meterLevel _ 0.
	meterRead _ true.
	soundPlaying _ nil.
	currentBuffer _ nil.
	nextIndex _ 1.
! !


!SoundRecorder methodsFor: 'accessing' stamp: 'jm 11/16/2001 22:09'!
closeFile
	"Close my output file."

	audioFile ifNotNil: [audioFile closeFile].
	audioFile _ nil.
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 11/20/2003 08:42'!
codec: aSoundCodec

	codec _ aSoundCodec.
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jj 10/20/97 15:30'!
isActive
	"Return true if I have a recordProcess running."

	^ recordProcess ~~ nil
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 9/2/97 16:16'!
isPaused
	"Return true if recording is paused."

	^ paused
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 5/12/2006 09:00'!
meterLevel
	"Return the meter level, an integer in the range [0..100] where zero is silence and 100 represents the maximum signal level possible without clipping."

	meterRead _ true.
	^ (100 * meterLevel) // 32768
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 7/4/1998 15:03'!
recordLevel

	^ recordLevel
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 7/4/1998 15:04'!
recordLevel: level
	"Set the desired recording level to the given value in the range 0.0 to 1.0, where 0.0 is the lowest recording level and 1.0 is the maximum. Do nothing if the sound input hardware does not support changing the recording level."
	"Details: On the Macintosh, the lowest possible record level attenuates the input signal, but does not silence it entirely." 

	recordLevel _ (level asFloat min: 1.0) max: 0.0.
	recordProcess ifNotNil: [
		self primSetRecordLevel: (1000.0 * recordLevel) asInteger].
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 11/16/2001 22:13'!
recordToFileNamed: fileName
	"Begin recording to the file with the given name."

	self closeFile.  "close previous file, if any"
	audioFile _ SunAudioFileWriter onFileNamed: fileName.
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'jm 12/15/97 14:28'!
samplingRate

	^ samplingRate
! !

!SoundRecorder methodsFor: 'accessing' stamp: 'di 2/16/1999 09:58'!
samplingRate: newRate

	samplingRate _ newRate  "Best are 44100 22050 11025"
! !


!SoundRecorder methodsFor: 'recording controls' stamp: 'di 2/17/1999 10:54'!
clearRecordedSound
	"Clear the sound recorded thus far. Go into pause mode if currently recording."

	paused _ true.
	recordedSound _ SequentialSound new.
	self allocateBuffer.
! !

!SoundRecorder methodsFor: 'recording controls' stamp: 'jm 10/20/2001 16:53'!
pause
	"Go into pause mode. The record level continues to be updated, but no sound is recorded."

	paused _ true.
	((currentBuffer ~~ nil) and: [nextIndex > 1])
		ifTrue: [self emitPartialBuffer.
				self allocateBuffer].

	soundPlaying ifNotNil: [
		soundPlaying pause.
		soundPlaying _ nil].

	self stopRecording.
! !

!SoundRecorder methodsFor: 'recording controls' stamp: 'jm 6/15/2003 23:54'!
playback
	"Playback the sound that has been recorded."

	self pause.
	soundPlaying _ self recordedSound.
	soundPlaying ifNotNil: [soundPlaying play].
! !

!SoundRecorder methodsFor: 'recording controls' stamp: 'jm 12/2/2003 20:55'!
resumeRecording
	"Resume recording after being paused."

	audioFile ifNotNil: [
		audioFile
			writeHeaderSamplingRate: samplingRate
			format: (SunAudioFileWriter formatCodeForCompressionType: #none)].
	self startRecording.
	paused _ false.
! !

!SoundRecorder methodsFor: 'recording controls' stamp: 'di 3/4/1999 22:38'!
startRecording
	"Turn of the sound input driver and start the recording process. Initially, recording is paused."

	| semaIndex |
	recordLevel ifNil: [recordLevel _ 0.5].  "lazy initialization"
	CanRecordWhilePlaying ifFalse: [SoundPlayer shutDown].
	recordProcess ifNotNil: [self stopRecording].
	paused _ true.
	meteringBuffer _ SoundBuffer newMonoSampleCount: 1024.
	meterLevel _ 0.
	self allocateBuffer.
	bufferAvailableSema _ Semaphore new.
	semaIndex _ Smalltalk registerExternalObject: bufferAvailableSema.
	self primStartRecordingDesiredSampleRate: samplingRate asInteger
		stereo: stereo
		semaIndex: semaIndex.
	samplingRate _ self primGetActualRecordingSampleRate.
	self primSetRecordLevel: (1000.0 * recordLevel) asInteger.
	recordProcess _ [self recordLoop] newProcess.
	recordProcess priority: Processor userInterruptPriority.
	recordProcess resume.
! !

!SoundRecorder methodsFor: 'recording controls' stamp: 'di 2/16/1999 09:13'!
stopRecording
	"Stop the recording process and turn of the sound input driver."

	recordProcess ifNotNil: [recordProcess terminate].
	recordProcess _ nil.
	self primStopRecording.
	Smalltalk unregisterExternalObject: bufferAvailableSema.
	((currentBuffer ~~ nil) and: [nextIndex > 1])
		ifTrue: [self emitPartialBuffer].
	self initializeRecordingState.
! !


!SoundRecorder methodsFor: 'trimming' stamp: 'jm 10/17/97 17:43'!
copyFrom: startPlace to: endPlace normalize: nFactor dcOffset: dcOffset
	"Return a new SoundBuffer containing the samples in the given range."

	| startBufIndex startSampleIndex endBufIndex endSampleIndex
	 count resultBuf j buf firstInBuf n |
	startBufIndex _ startPlace at: 1.
	startSampleIndex _ startPlace at: 2.
	endBufIndex _ endPlace at: 1.
	endSampleIndex _ endPlace at: 2.

	startBufIndex = endBufIndex
		ifTrue: [count _ endSampleIndex + 1 - startSampleIndex]
		ifFalse: [
			count _ ((recordedBuffers at: startBufIndex) size + 1 - startSampleIndex).  "first buffer"
			count _ count + endSampleIndex.  "last buffer"
			startBufIndex + 1 to: endBufIndex - 1 do:
				[:i | count _ count + (recordedBuffers at: i) size]].  "middle buffers"
	resultBuf _ SoundBuffer newMonoSampleCount: count.

	j _ 1.  "next destination index in resultBuf"
	startBufIndex to: endBufIndex do: [:i |
		buf _ recordedBuffers at: i.
		firstInBuf _ 1.
	 	n _ buf size.
		i = startBufIndex ifTrue: [
			n _ (recordedBuffers at: startBufIndex) size + 1 - startSampleIndex.
			firstInBuf _ startSampleIndex].
		i = endBufIndex ifTrue: [
			i = startBufIndex
				ifTrue: [n _ endSampleIndex + 1 - startSampleIndex]
				ifFalse: [n _ endSampleIndex]].
		self copyTo: resultBuf from: j to: (j + n - 1)
			from: buf startingAt: firstInBuf
			normalize: nFactor dcOffset: dcOffset.
		j _ j + n].
	^ resultBuf
! !

!SoundRecorder methodsFor: 'trimming' stamp: 'jm 10/18/97 11:23'!
copyTo: resultBuf from: startIndex to: endIndex from: buf startingAt: firstInBuf normalize: nFactor dcOffset: dcOffset
	"Copy samples from buf to resultBuf removing the DC offset and normalizing their volume in the process."

	| indexOffset |
	indexOffset _ firstInBuf - startIndex.
	startIndex to: endIndex do: [:i |
		resultBuf at: i put: (((buf at: (i + indexOffset)) - dcOffset) * nFactor) // 1000].
! !

!SoundRecorder methodsFor: 'trimming' stamp: 'di 2/16/1999 22:11'!
endPlace

	^ Array with: recordedBuffers size with: recordedBuffers last size! !

!SoundRecorder methodsFor: 'trimming' stamp: 'di 2/16/1999 22:11'!
firstSampleOverThreshold: threshold dcOffset: dcOffset startingAt: startPlace
	"Beginning at startPlace, this routine will return the first place at which a sample exceeds the given threshold."

	| buf s iStart jStart nThreshold |
	nThreshold _ threshold negated.
	iStart _ startPlace first.
	jStart _ startPlace second.
	iStart to: recordedBuffers size do:
		[:i | buf _ recordedBuffers at: i.
		jStart to: buf size do:
			[:j | s _ (buf at: j) - dcOffset.
			(s < nThreshold or: [s > threshold]) ifTrue:
				["found a sample over threshold"
				^ Array with: i with: j]].
		jStart _ 1].
	^ self endPlace! !

!SoundRecorder methodsFor: 'trimming' stamp: 'jm 10/18/97 11:22'!
normalizeFactorFor: percentOfMaxVolume min: min max: max dcOffset: dcOffset
	"Return a normalization factor for the range of sample values and DC offset. A normalization factor is a fixed-point number that will be divided by 1000 after multiplication with each sample value."

	| peak factor |
	peak _ (max - dcOffset) max: (min - dcOffset) negated.
	peak = 0 ifTrue: [^ 1000].
	factor _ (32767.0 * percentOfMaxVolume) / (100.0 * peak).
	^ (factor * 1000.0) asInteger
! !

!SoundRecorder methodsFor: 'trimming' stamp: 'di 2/16/1999 23:01'!
place: startPlace plus: nSamples
	"Return the place that is nSamples (may be negative) beyond thisPlace."

	| i j remaining buf |
	i _ startPlace first.
	j _ startPlace second.
	nSamples >= 0
	ifTrue: [remaining _ nSamples.
			[buf _ recordedBuffers at: i.
			(j + remaining) <= buf size ifTrue: [^ Array with: i with: j + remaining].
			i < recordedBuffers size]
				whileTrue: [remaining _ remaining - (buf size - j + 1).
							i _ i+1.  j _ 1].
			^ self endPlace]
	ifFalse: [remaining _ nSamples negated.
			[buf _ recordedBuffers at: i.
			(j - remaining) >= 1 ifTrue: [^ Array with: i with: j - remaining].
			i > 1]
				whileTrue: [remaining _ remaining - j.
							i _ i-1.  j _ (recordedBuffers at: i) size].
			^ #(1 1)]! !

!SoundRecorder methodsFor: 'trimming' stamp: 'di 2/16/1999 23:19'!
scanForEndThreshold: threshold dcOffset: dcOffset minLull: lull startingAt: startPlace
	"Beginning at startPlace, this routine will find the last sound that exceeds threshold, such that if you look lull samples later you will not find another sound over threshold within the following block of lull samples.
	Return the place that is lull samples beyond to that last sound.
	If no end of sound is found, return endPlace."

	| buf s iStart jStart nThreshold n |
	nThreshold _ threshold negated.
	iStart _ startPlace first.
	jStart _ startPlace second.
	n _ 0.
	iStart to: recordedBuffers size do:
		[:i | buf _ recordedBuffers at: i.
		jStart to: buf size do:
			[:j | s _ (buf at: j) - dcOffset.
			(s < nThreshold or: [s > threshold])
				ifTrue: ["found a sample over threshold"
						n _ 0]
				ifFalse: ["still not over threshold"
						n _ n + 1.
						n >= lull ifTrue: [^ Array with: i with: j]]].
		jStart _ 1].
	^ self endPlace! !

!SoundRecorder methodsFor: 'trimming' stamp: 'jm 6/26/2005 12:06'!
scanForEndThreshold: threshold dcOffset: dcOffset startingAt: startPlace
	"Scan backward from the end to find the last buffer that has a sample over the given threshold. If no such sample is is found, return startPlace."

	| nThreshold buf lastIndex s |
	nThreshold _ threshold negated.
	recordedBuffers size to: startPlace first by: -1 do: [:i |
		buf _ recordedBuffers at: i.
		lastIndex _ buf monoSampleCount.
		"ignore the last little bit of the last buffer--it could be just the mouse click to stop recording:"
		i = recordedBuffers size ifTrue: [lastIndex _ (lastIndex - (samplingRate // 3)) max: 1].
		1 to: lastIndex do: [:j |
			s _ (buf at: j) - dcOffset.
			(s < nThreshold or: [s > threshold]) ifTrue: [
				^ Array with: i with: lastIndex]]].
	^ startPlace
! !

!SoundRecorder methodsFor: 'trimming' stamp: 'di 3/4/1999 22:13'!
scanForStartThreshold: threshold dcOffset: dcOffset minDur: duration startingAt: startPlace
	"Beginning at startPlace, this routine will find the first sound that exceeds threshold, such that if you look duration samples later you will find another sound over threshold within the following block of duration samples.
	Return the place that is duration samples prior to that first sound.
	If no sound is found, return endPlace."

	| soundPlace lookPlace nextSoundPlace thirdPlace |
	soundPlace _ self firstSampleOverThreshold: threshold dcOffset: dcOffset
					startingAt: startPlace.
	[soundPlace = self endPlace ifTrue: [^ soundPlace].
	"Found a sound -- look duration later"
	lookPlace _ self place: soundPlace plus: duration.
	nextSoundPlace _ self firstSampleOverThreshold: threshold dcOffset: dcOffset
					startingAt: lookPlace.
	thirdPlace _ self place: lookPlace plus: duration.
	nextSoundPlace first < thirdPlace first
		or: [nextSoundPlace first = thirdPlace first
			and: [nextSoundPlace second < thirdPlace second]]]
		whileFalse: [soundPlace _ nextSoundPlace].

	"Yes, there is sound in the next interval as well"
	^ self place: soundPlace plus: 0-duration
! !

!SoundRecorder methodsFor: 'trimming' stamp: 'jm 6/26/2005 11:31'!
trim: threshold normalizedVolume: percentOfMaxVolume
	"Remove the leading and trailing parts of this recording that are below the given threshold. Remove any DC offset and scale the recording so that its peaks are the given percent of the maximum volume."

	| max min sum totalSamples bufSize s dcOffset startPlace endPlace resultBuf nFactor |
	recordedSound ifNil: [^ self].
	stereo ifTrue: [self error: 'stereo trimming is not yet supported'].
	paused ifFalse: [self error: 'must stop recording before trimming'].
	recordedBuffers _ recordedSound sounds collect: [:snd | snd samples].
	recordedBuffers isEmpty ifTrue: [^ self].

	max _ min _ sum _ totalSamples _ 0.
	recordedBuffers do: [:buf |
		bufSize _ buf size.
		totalSamples _ totalSamples + buf size.
		1 to: bufSize do: [:i |
			s _ buf at: i.
			s > max ifTrue: [max _ s].
			s < min ifTrue: [min _ s].
			sum _ sum + s]].
	dcOffset _ sum // totalSamples.

	"a place is an array of <buffer index><index of sample in buffer>"
	startPlace _ self scanForStartThreshold: threshold
					dcOffset: dcOffset
					minDur: (samplingRate // 200)
					startingAt: #(1 1).
	startPlace = self endPlace ifTrue: [
		"no samples above threshold"
		recordedBuffers _ nil.
		^ self clearRecordedSound].

	endPlace _ self scanForEndThreshold: threshold
					dcOffset: dcOffset
					startingAt: startPlace.
	nFactor _ self normalizeFactorFor: percentOfMaxVolume min: min max: max dcOffset: dcOffset.
	startPlace _ self place: startPlace plus: (samplingRate // 5) negated.  "back up a bit"
	resultBuf _ self copyFrom: startPlace to: endPlace normalize: nFactor dcOffset: dcOffset.
	recordedSound _ SampledSound new setSamples: resultBuf samplingRate: samplingRate.
	recordedBuffers _ nil.
! !


!SoundRecorder methodsFor: 'results' stamp: 'jm 6/15/2003 23:56'!
condensedSamples
	"Return a single SoundBuffer that is the contatenation of all my recorded buffers."

	| sz newBuf i |
	recordedSound ifNil: [^ SoundBuffer new: 0].
	recordedBuffers _ recordedSound sounds collect: [:snd | snd samples].
	recordedBuffers isEmpty ifTrue: [^ SoundBuffer new: 0].
	recordedBuffers size = 1 ifTrue: [^ recordedBuffers first copy].
	sz _ recordedBuffers inject: 0 into: [:tot :buff | tot + buff size].
	newBuf _ SoundBuffer newMonoSampleCount: sz.
	i _ 1.
	recordedBuffers do: [:b |
		1 to: b size do: [:j |
			newBuf at: i put: (b at: j).
			i _ i + 1]].
	recordedBuffers _ nil.
	^ newBuf
! !

!SoundRecorder methodsFor: 'results' stamp: 'di 2/16/1999 20:49'!
condensedStereoSound
	"Decompose my buffers into left and right channels and return a mixed sound consisting of the those two channels. This may be take a while, since the data must be copied into new buffers."

	| sz leftBuf rightBuf leftI rightI left |
	sz _ recordedBuffers inject: 0 into: [:tot :buff | tot + buff size].
	leftBuf _ SoundBuffer newMonoSampleCount: (sz + 1) // 2.
	rightBuf _ SoundBuffer newMonoSampleCount: (sz + 1) // 2.
	leftI _ rightI _ 1.
	left _ true.
	recordedBuffers do: [:b |
		1 to: b size do: [:j |
			left
				ifTrue: [leftBuf at: leftI put: (b at: j). leftI _ leftI + 1. left _ false]
				ifFalse: [rightBuf at: rightI put: (b at: j). rightI _ rightI + 1. left _ true]]].
	^ MixedSound new
		add: (SampledSound new setSamples: leftBuf samplingRate: samplingRate) pan: 0.0;
		add: (SampledSound new setSamples: rightBuf samplingRate: samplingRate) pan: 1.0
! !

!SoundRecorder methodsFor: 'results' stamp: 'di 2/17/1999 11:07'!
recordedSound
	"Return the sound that was recorded."

	^ recordedSound
! !


!SoundRecorder methodsFor: 'multitrack' stamp: 'jm 12/12/2001 10:47'!
estimateRecordPlaySampleOffset
	"Measure the number of samples of lag between the sound being played and the sound being recorded. Collect and answer 10 measurements. If there is little variation among these measurements, your sound hardware and OS can probably be used for multi-track recording. If there is wide variation, it may merely be that the record level needs to be adjusted. You can get a sense of this by setting viewBuffer to true to view the buffer from one test. If you are getting a clean recording with a strong (perhaps even clipped) signal and noise level well below the threshold, then the variation is due to variability in the timing of the OS sound driver startup sequence, and synchronization amoung tracks may be poor."
	"Details: This estimate is made by playing a short pulse while recording, then finds the beginning of the pulse in the recorded sound. For this to work, the sound output must be fed into the input. The simplest way to do this is to place the recording microphone near the speaker and remain quiet while the test is run. You can also use a patch cord to connect the sound output to the input directly, in which case silence is not necessary."

	| threshold recLevel viewBuffer rec snd buf |
	threshold _ 10000.
	recLevel _ 0.5.  "it may be necessary to adjust this to get a clean recording"
	viewBuffer _ false.  "set this true to look at the buffer collected"
	^ (1 to: 10) collect: [:i |
		(Delay forMilliseconds: (1 to: 200) atRandom) wait.
		rec _ SoundRecorder new.
		rec recordLevel: recLevel.
		snd _ FMSound new setPitch: 1000.0 dur: 0.005 loudness: 0.5.
		rec startPlayingAndRecording.
		snd play.
		(Delay forMilliseconds: 500) wait.
		rec stopRecording.
		buf _ rec condensedSamples.
		viewBuffer ifTrue: [WaveEditor openOn: buf. ^ true].
		buf indexOfFirstSampleOver: threshold]
! !

!SoundRecorder methodsFor: 'multitrack' stamp: 'jm 12/18/2001 08:14'!
mixed: trackList
	"Answer a Mixed sound that mixes the given list of sounds. The track list contains SoundBuffers which are assumed to be monophonic sounds sampled at my sampling rate."

	| mix snd |
	self stopRecording; clearRecordedSound.
	mix _ MixedSound new.
	trackList do: [:buf |
		snd _ SampledSound samples: buf samplingRate: samplingRate.
		mix add: snd].
	^ mix
! !

!SoundRecorder methodsFor: 'multitrack' stamp: 'jm 12/12/2001 12:21'!
recordWhilePlaying: trackList offsetSampleCount: offsetSampleCount
	"Record while playing the given set of tracks. Each track is a monophonic SoundBuffer. These are played as a bunch of LoopedSampledSounds mixed together. The offsetSampleCount parameter should be the average of the sample offsets measured by doing:
	SoundRecorder new estimateRecordPlaySampleOffset"

	| mixer buf |
	mixer _ self mixed: trackList.
	self startPlayingAndRecording.
	mixer play.
	self inform: 'Recording. Click OK to end recording'.
	self stopRecording.
	buf _ self condensedSamples.
	^ buf copyFrom: offsetSampleCount to: buf size
! !

!SoundRecorder methodsFor: 'multitrack' stamp: 'jm 12/12/2001 14:47'!
startPlayingAndRecording
	"Start both the sound player and sound recorder so that (with luck) the time offset between the outgoing and incoming sample streams will be the same every time (with the same hardware and system configuration)."
	"Note: Sets CanRecordWhilePlaying to true. Be sure your recording hardware can support simultaneous record and playback."

	| rest restCount |
	self class canRecordWhilePlaying ifFalse: [
		^ self inform:
			'Set the canRecordWhilePlaying preference to true if your hardware supports it'].

	"first, shut down the sound player and recorder"
	self stopRecording; clearRecordedSound.
	SoundPlayer shutDown.

	"clean up any garbage before starting"
	Smalltalk garbageCollect.

	"start the sound player and wait for the first interrupt"
	rest _ RestSound dur: 0.1.
	restCount _ rest samplesRemaining.
	SoundPlayer startUp.
	SoundPlayer resumePlaying: rest quickStart: false.
	[rest samplesRemaining = restCount] whileTrue.  "wait for a buffer flip interrupt"
	(Delay forMilliseconds: 5) wait.  "wait a bit longer"

	"start the recorder"
	self resumeRecording.
! !


!SoundRecorder methodsFor: 'primitives' stamp: 'bf 5/11/2012 17:20'!
primGetActualRecordingSampleRate
	"Return the actual sample rate being used for recording. This primitive fails unless sound recording is currently in progress."

	<primitive: 'primitiveSoundGetRecordingSampleRate' module: 'SoundPlugin'>
	self primitiveFailed
! !

!SoundRecorder methodsFor: 'primitives' stamp: 'bf 5/11/2012 17:20'!
primRecordSamplesInto: aWordArray startingAt: index
	"Record a sequence of 16-bit sound samples into the given array starting at the given sample index. Return the number of samples recorded, which may be zero if no samples are currently available."

	<primitive: 'primitiveSoundRecordSamples' module: 'SoundPlugin'>
	self primitiveFailed
! !

!SoundRecorder methodsFor: 'primitives' stamp: 'bf 5/11/2012 17:20'!
primSetRecordLevel: anInteger
	"Set the desired recording level to the given value in the range 0-1000, where 0 is the lowest recording level and 1000 is the maximum. Do nothing if the sound input hardware does not support changing the recording level."

	<primitive: 'primitiveSoundSetRecordLevel' module: 'SoundPlugin'>
	self primitiveFailed
! !

!SoundRecorder methodsFor: 'primitives' stamp: 'bf 5/11/2012 17:20'!
primStartRecordingDesiredSampleRate: samplesPerSec stereo: stereoFlag semaIndex: anInteger
	"Start sound recording with the given stereo setting. Use a sampling rate as close to the desired rate as the underlying platform will support. If the given semaphore index is > 0, it is taken to be the index of a Semaphore in the external objects array to be signalled every time a recording buffer is filled."

	<primitive: 'primitiveSoundStartRecording' module: 'SoundPlugin'>
	self primitiveFailed
! !

!SoundRecorder methodsFor: 'primitives' stamp: 'bf 5/11/2012 17:20'!
primStopRecording
	"Stop sound recording. Does nothing if recording is not currently in progress."

	<primitive: 'primitiveSoundStopRecording' module: 'SoundPlugin'>
	self primitiveFailed
! !


!SoundRecorder methodsFor: 'private' stamp: 'di 2/17/1999 11:13'!
allocateBuffer
	"Allocate a new buffer and reset nextIndex."

	| bufferTime |
	bufferTime _ stereo  "Buffer time = 1/2 second"
		ifTrue: [self samplingRate asInteger]
		ifFalse: [self samplingRate asInteger // 2].
	currentBuffer _ SoundBuffer newMonoSampleCount:
		"Multiple of samplesPerFrame that is approx. bufferTime long"
		(bufferTime truncateTo: self samplesPerFrame).
	nextIndex _ 1.
! !

!SoundRecorder methodsFor: 'private' stamp: 'jm 11/20/2003 08:42'!
emitBuffer: buffer

	| sound |
	audioFile ifNotNil: [  "recording to file"
		audioFile appendSamples: buffer.
		^ self].

	sound _ SampledSound new setSamples: buffer samplingRate: samplingRate.
	recordedSound add:
		(codec == nil
			ifTrue: [sound]
			ifFalse: [codec compressSound: sound]).
! !

!SoundRecorder methodsFor: 'private' stamp: 'di 2/17/1999 11:13'!
emitPartialBuffer
	| s |
	s _ self samplesPerFrame.
	self emitBuffer: (currentBuffer copyFrom: 1 to: ((nextIndex-1) +( s-1) truncateTo: s))! !

!SoundRecorder methodsFor: 'private' stamp: 'jm 5/12/2006 09:25'!
meterFrom: start count: count in: buffer
	"Update the meter level with the maximum signal level in the given range of the given buffer."

	| last max sample |
	count = 0 ifTrue: [^ self].  "no new samples"
	last _ start + count - 1.
	max _ 0.
	start to: last do: [:i |
		sample _ buffer at: i.
		sample < 0 ifTrue: [sample _ sample negated].
		sample > max ifTrue: [max _ sample]].
	meterRead
		ifTrue: [meterLevel _ max]
		ifFalse: [meterLevel _ meterLevel max: max].  "if the meter level hasn't been read, remember maximum"
	meterRead _ false.
! !

!SoundRecorder methodsFor: 'private' stamp: 'di 2/16/1999 08:55'!
recordLoop
	"Record process loop that records samples."

	| n sampleCount |
	n _ 0.
	[true] whileTrue: [
		n = 0 ifTrue: [bufferAvailableSema wait].
		paused
			ifTrue: [
				n _ self primRecordSamplesInto: meteringBuffer startingAt: 1.
				self meterFrom: 1 count: n in: meteringBuffer]
			ifFalse: [
				n _ self primRecordSamplesInto: currentBuffer startingAt: nextIndex.
				self meterFrom: nextIndex count: n in: currentBuffer.
				nextIndex _ nextIndex + n.
				stereo
					ifTrue: [sampleCount _ currentBuffer stereoSampleCount]
					ifFalse: [sampleCount _ currentBuffer monoSampleCount].
				nextIndex > sampleCount
					ifTrue: [
						self emitBuffer: currentBuffer.
						self allocateBuffer]]].
! !

!SoundRecorder methodsFor: 'private' stamp: 'jm 11/20/2003 08:43'!
samplesPerFrame
	"Can be overridden to quantize buffer size for, eg, fixed-frame codecs"

	codec == nil
		ifTrue: [^ 1]
		ifFalse: [^ codec samplesPerFrame].
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SoundRecorder class
	instanceVariableNames: ''!

!SoundRecorder class methodsFor: 'class initialization' stamp: 'jm 11/20/2003 08:38'!
initialize
	"SoundRecorder initialize"
	"Details: Some computers cannot record and playback sound at the same time. If CanRecordWhilePlaying is false, then the SoundRecorder alternates between recording and playing. If it is true, sounds can be playing during recording."

	CanRecordWhilePlaying _ true.
! !


!SoundRecorder class methodsFor: 'accessing' stamp: 'jj 10/20/97 15:32'!
canRecordWhilePlaying
	"Return true if this platform supports simultaneous sound recording and playback."

	^ CanRecordWhilePlaying
! !


SoundRecorder initialize!
