EnchantingDeviceConfiguration subclass: #NXTConfiguration
	instanceVariableNames: 'sensors motors pilot '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Enchanting-Configuration'!
!NXTConfiguration commentStamp: '<historical>' prior: 0!
NXTConfiguration contains all the information necessary to configure a LEGO Mindstorms NXT device (by specifying which sensors are in use and where they are attached.)!


!NXTConfiguration methodsFor: 'accessing' stamp: 'CWB 12/15/2010 22:15'!
configuredPeripherals
"Returns a list of all the configured peripherals"
	| list |

	list _ OrderedCollection new.
	list addAll: motors.
	list addAll: sensors.
	list add: pilot.
	"Might add things that are always configured here."

	list _ list reject: [:item | item isNil].
	^ list.
! !

!NXTConfiguration methodsFor: 'accessing' stamp: 'CWB 1/17/2013 10:53'!
setOfConfiguredClasses
"Returns a set of all the configured classes"

	| list |

	list _ OrderedCollection new.
	list _ self configuredPeripherals collect: [:item | item class].
	list add: NXTButtons.	"There is always one set of buttons (under the display)."
	list add: NXTRam.	"The memory device is always available, too."
	^ list asSet.! !


!NXTConfiguration methodsFor: 'dialog interaction' stamp: 'CWB 12/20/2012 11:19'!
applyChangesFrom: dialog named: dialogName over: scratchFrame
"The user pressed okay on the dialog; update structures to reflect the changes."

	| destArray nameMapping |

	destArray _ nil.

	"In this map, we will take previous names and use them as keys, setting all values to nil.  Then we will look through all peripherals, put their original name in as a key, and their new name in as a value.  Anything that still has 'nil' as a value has been deleted."
	nameMapping _ Dictionary new.

	(dialogName = 'sensors') ifTrue: [ destArray _ sensors ].
	(dialogName = 'motors') ifTrue: [ destArray _ motors ].
	(dialogName = 'driving') ifTrue: [ destArray _ OrderedCollection with: pilot ].

	destArray ifNil: [ self error: 'Unknown configuration dialog' ].

	"Get the previous names"
	destArray do: [:p |
		p ifNotNil: [ nameMapping at: (p name) put: nil ].
	].


	(1 to: destArray size) do: [ :index |
		| p |
		p _ dialog peripheralAt: index.
		destArray at: index put: p.
		p ifNotNil: [ 
			| configMorph |
			p portIndex: index.
			configMorph _ dialog blockAt: index.
			nameMapping at: (configMorph initialName) put: (p name).
		 ].
	].

	(dialogName = 'driving') ifTrue: [ pilot _ destArray first ].

	"Update any names that have changed - unless we are configuring the pilot, which has no name."
	(dialogName = 'driving') ifFalse: [
		self replaceNamesInScriptsUsing: nameMapping over: scratchFrame.
	].
! !

!NXTConfiguration methodsFor: 'dialog interaction' stamp: 'CWB 12/20/2012 11:20'!
portsFor: dialogName
"Returns an array containing:
	an array with the names of all the ports
	an array (of the same size) of ConfigurationBlockMorphs (or nil) for each port."

	| sourceArray baseName count countMethod portNames portBlocks |

	sourceArray _ nil.

	(dialogName = 'sensors') ifTrue: [
		sourceArray _ sensors.
		baseName _ 'Sensor Port'.
		countMethod _ #numeric.		
	].
	(dialogName = 'motors') ifTrue: [
		sourceArray _ motors.
		baseName _ 'Motor Port'.
		countMethod _ #alphabetical.		
	].
	(dialogName = 'driving') ifTrue: [
		sourceArray _ OrderedCollection with: pilot.
		baseName _ 'Drive Type'.
		countMethod _ #none.		
	].

	portNames _ OrderedCollection new.
	portBlocks _ OrderedCollection new.

	sourceArray ifNil: [ self beep. ^ { portNames . portBlocks } ].

	count _ 1.
	sourceArray do: [ :peripheral |
		portNames add: (self class portName: baseName at: count method: countMethod).
		portBlocks add: (peripheral ifNotNil: [peripheral createConfigBlock] ifNil: [nil]).
		count _ count + 1.
	].
		

	^ { portNames . portBlocks }! !


!NXTConfiguration methodsFor: 'initialization' stamp: 'CWB 12/17/2010 15:08'!
initialize

	super initialize.
	motors _ OrderedCollection with: nil with: nil with: nil.
	sensors _ OrderedCollection with: nil with: nil with: nil with: nil.
	pilot _ nil.
! !


!NXTConfiguration methodsFor: 'object i/o' stamp: 'CWB 12/7/2010 20:35'!
fieldsVersion

	^ 1! !

!NXTConfiguration methodsFor: 'object i/o' stamp: 'CWB 12/7/2010 20:40'!
initFieldsFrom: anObjStream version: classVersion

	super initFieldsFrom: anObjStream version: classVersion.
	self initFieldsNamed: #(
		sensors
		motors
		pilot
	) from: anObjStream.! !

!NXTConfiguration methodsFor: 'object i/o' stamp: 'CWB 12/7/2010 20:38'!
storeFieldsOn: anObjStream

	super storeFieldsOn: anObjStream.
	self storeFieldsNamed: #(
		sensors
		motors
		pilot
	) on: anObjStream.! !


!NXTConfiguration methodsFor: 'other' stamp: 'CWB 11/21/2011 17:29'!
customExportColor: aColor for: aBlock on: aStream with: anExporter

"Outputs something like 'EObject.createList(240.0f, 12.0f, 38.0f)' where the numbers are:
  hue -- the color, 0 to 360 (degrees),
  saturation -- how true/vivid the colour is, 0 to 100 (%),
  brightness -- how bright the colour is, 0 to 100 (%)"

	aStream nextPutAll: 'EObject.createList('.
	anExporter exportLiteralValue: ((aColor hue) roundTo: 0.1) on: aStream.
	aStream nextPutAll: ', '.
	anExporter exportLiteralValue: ((aColor saturation * 100) roundTo: 0.1) on: aStream.
	aStream nextPutAll: ', '.
	anExporter exportLiteralValue: ((aColor brightness * 100) roundTo: 0.1) on: aStream.
	aStream nextPutAll: ')'.


! !

!NXTConfiguration methodsFor: 'other' stamp: 'CWB 11/20/2012 21:24'!
uncoloredArgMorphFor: specString

	"Allows the NXT device to create an argument other than that created by CommandBlockMorph."

	"Returns nil or an argMorph."

	| code |
	
	code _ specString at: 2.

	"instrument"
	$I = code ifTrue: [^ 
		ExpressionArgMorphWithMenu new numExpression: '1';
		menuSelector: #nxtInstrumentMenu ].

	"Steering rate"
	$q = code ifTrue: [^ 
		ExpressionArgMorphWithMenu new numExpression: '100';
		menuSelector: #nxtSteeringMenu ].

	^ nil.
! !


!NXTConfiguration methodsFor: 'export sprite' stamp: 'CWB 7/16/2012 18:43'!
arrangeDataOfCostume: aCostume discriminatorBlock: aBlock
"Returns a byte array with all the image data from a sprite's costume, in the right format for the NXT"
"This code is ported from NXJImageConverter.java function nxjImageConvert"

	| w h n data index color form |

	form _ aCostume compositeForm.
	w _ form width.
	h _ form height.

	n _ h // 8.  "8 pixels per byte"
	(h > (n * 8)) ifTrue: [ n _ n + 1 ].
	data _ ByteArray new: (w * n).

	index _ 1.

	1 to: h by: 8 do: [ :i | 
		1 to: w do: [ :j |
			| d |
			d _ 0.
			7 to: 0 by: -1 do: [ :k | 
				| x y bit |		
				d _ d bitShift: 1.
				x _ j - 1.
				y _ i + k - 1.
				(y < h) ifTrue: [
					color _ form colorAt: x@y.
					bit _ aBlock value: color.
					d _ d bitOr: bit.
				].
			].
			data byteAt: index put: d.
			index _ index + 1.
		].
	].

	^ data.
! !

!NXTConfiguration methodsFor: 'export sprite' stamp: 'CWB 6/2/2012 15:03'!
arrangeDataOfCostumeImage: aCostume
"Returns a byte array with all the image data from a sprite's costume, in the right format for the NXT"

^ self arrangeDataOfCostume: aCostume 
	discriminatorBlock: [ :colour |
		(colour isTransparent) ifTrue: [ 0 ] ifFalse: [ colour closestPixelValue1 ].
	].

! !

!NXTConfiguration methodsFor: 'export sprite' stamp: 'CWB 6/2/2012 15:03'!
arrangeDataOfCostumeMask: aCostume
"Returns a byte array with all the image data from a sprite's costume, in the right format for the NXT"

^ self arrangeDataOfCostume: aCostume 
	discriminatorBlock: [ :colour |
		(colour isTransparent) ifTrue: [ 1 ] ifFalse: [ 0 ].
	].

! !

!NXTConfiguration methodsFor: 'export sprite' stamp: 'CWB 8/20/2012 06:08'!
exportCostume: aCostume on: aBinaryStream
"Outputs a representation of a costume onto a stream"

	| imageData maskData w h offsetX offsetY imageCount |

	imageData _ self arrangeDataOfCostumeImage: aCostume.
	maskData _ self arrangeDataOfCostumeMask: aCostume.
	w _ aCostume form width.
	h _ aCostume form height.
	offsetX _ aCostume rotationCenter x.
	offsetY _ aCostume rotationCenter y.

	"We will now proceed to write the data onto the stream."

	"Write out header information that is not a part of either image"
	imageCount _ 2.		"we have a mask and an image"
	aBinaryStream
		int16: imageCount;
		int16: offsetX;
		int16: offsetY. 

	"Write out the mask image"	
	aBinaryStream nextPutAll: 'LNI0' asByteArray;	"sentinel value"
		uint16: w;
		uint16: h;
		nextPutAll: maskData.

	"Write out the image's image"
	aBinaryStream nextPutAll: 'LNI0' asByteArray;
		uint16: w;
		uint16: h;
		nextPutAll: imageData.
! !

!NXTConfiguration methodsFor: 'export sprite' stamp: 'CWB 11/20/2012 11:41'!
exportCostumesToFileNamed: filename for: aStageOrSprite

	"This function writes a file with all the data for all the costumes,
	and header information with indexes to the starting points of the costumes."

	"The file will look like this:
		SPR0 - sentinel value, 4 bytes
		number of costumes, 2 bytes
		starting position (file offset) for costume 1 data, 4 bytes
		starting position for costume 2, 3, ... n, 4 bytes each
		name of costume 1, 2, 3, ... n.   (2 bytes for length, followed by UTF data)
		data for costume 1, 2, 3 ... n"
	
	| f numCostumes nameData costumeData headerSize offset |

	f _ StandardFileStream newScratchFileNamed: filename.
	f ifNil: [^ self].  "could not create or open file for writing"
	f binary.


	"Go through each costume and collect the names and image data."
	numCostumes _ aStageOrSprite numberOfCostumes.
	nameData _ WriteStream on: (Array new: 1024).
	costumeData _ OrderedCollection new.
	(1 to: numCostumes) do: [ 
		:index | | name data |

		name _ (aStageOrSprite costumeNameFromNumber: index) asUTF8.
		nameData string: name.

		data _ WriteStream on: (Array new: 2048).
		self exportCostume: (aStageOrSprite costumeAtIndex: index) on: data.
		costumeData add: data.
	].


	"Write the data"
	f nextPutAll: 'SPR0' asByteArray.
	f uint16: numCostumes.
	
	"Data file position offsets."
	headerSize _ 4 + 2 + (4 * numCostumes) + nameData size.
	offset _ headerSize.
	(1 to: numCostumes) do: [ :i |
		f int32: offset.
		offset _ offset + (costumeData at: i) size.
	].

	f nextPutAll: nameData contents asByteArray.

	(1 to: numCostumes) do: [ :i |
		f nextPutAll: (costumeData at: i) contents asByteArray.
	].

	f close.

! !

!NXTConfiguration methodsFor: 'export sprite' stamp: 'CWB 9/7/2012 06:20'!
exportSprite: aSprite on: aStream for: anExporter

	| position baseFilename filename numCostumes currentCostume direction hidden ml |

 
	position _ aSprite referencePosition.
	(aSprite isKindOf: ScratchStageMorph) 
		ifTrue: [ position _ 0@0 ].
	
	"The constructor we are targetting is:
	public EncSprite(float x, float y, String baseFilename, int numCostumes, int currentCostume)".

	baseFilename _ aSprite exportedSpritePrefix, '.cstm'.
	filename _ (ScratchFileChooserDialog userEnchantingTempDir) fullNameFor: baseFilename.
	numCostumes _ aSprite numberOfCostumes.
	currentCostume _ aSprite costumeIndex.
	direction _ aSprite rotationDegrees.
	hidden _ aSprite isHidden.


	"Output used to be:"

	"sprite_name = new EncSprite(10, 20, new Costume[] { .... });"
	"Note that 'sprite_name = new EncSprite' and ');' are output by the template."

	anExporter exportLiteralValue: position x on: aStream.
	aStream nextPutAll: ', '.

	anExporter exportLiteralValue: position y on: aStream.
	aStream nextPutAll: ', '.

	anExporter exportLiteralValue: baseFilename on: aStream.
	aStream nextPutAll: ', '.

	aStream nextPutAll: numCostumes asString, ', '.
	aStream nextPutAll: currentCostume asString, ', '.

	anExporter exportLiteralValue: direction on: aStream.
	aStream nextPutAll: ', '.

	anExporter exportLiteralValue: hidden on: aStream.


	"Now, write a file with all the information for all the costumes"
	self exportCostumesToFileNamed: filename for: aSprite.



	"Add the filename to medialist.dat."
	ml := CrLfFileStream fileNamed: ((ScratchFileChooserDialog userEnchantingTempDir) fullNameFor: 'mediaList.dat').

	ml setToEnd.
	ml nextPutAll: baseFilename; cr.
	ml close.
! !


!NXTConfiguration methodsFor: 'display' stamp: 'CWB 7/17/2012 14:32'!
mapFromDisplayToScreenSpace: point
	"Take a point that is specified in the coordinate space for your display,
	and return a corresponding point relative to the center of the on-screen stage."

	^ point * 4 + (0@2).! !

!NXTConfiguration methodsFor: 'display' stamp: 'CWB 7/17/2012 14:33'!
mapFromScreenToDisplaySpace: point
	"Take a point that is specified relative to the center of the on-screen stage,
	and return the corresponding coordinate suitable for the device at hand."

	^ (point + (0@2)) / 4.! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

NXTConfiguration class
	instanceVariableNames: ''!

!NXTConfiguration class methodsFor: 'dialog interaction' stamp: 'CWB 11/20/2012 19:44'!
blockColorFor: dialogName
"Returns the color ConfigurationBlockMorphs should be in the specified dialog"
	
	| mapping category |

	mapping _ Dictionary newFrom: {'sensors' -> 'sensing'}.
	category _ mapping at: dialogName ifAbsent: [ dialogName ].	

	^ super blockColorFor: category.
! !

!NXTConfiguration class methodsFor: 'dialog interaction' stamp: 'CWB 1/17/2013 10:52'!
orderedPeripheralClasses

	"Returns a list of all peripherals that might be set up for this device.
	The order is important, as it will be followed when placing blocks in the
	palettes in the main user interface.  (In other words, if you return 
	{ WhizBang . SawBuzz } the WhizBang blocks will always appear before
	the SawBuzz blocks."

	^ { 
		NXTDifferentialPilot .
		NXTDCMotor . 
		NXTMotor .
		NXTButtons .
		NXTRam .
		NXTTouchSensor . 
		NXTRangeDetector . 
		NXTLightDetector . 
		NXTSoundSensor .
		NXTCompassSensor .
		NXTGyroSensor
	}! !

!NXTConfiguration class methodsFor: 'dialog interaction' stamp: 'CWB 12/20/2012 11:19'!
peripheralsFor: dialogName
"Returns a list of all the peripherals for the specified dialog"
	
	(dialogName = 'sensors') ifTrue: [
		^ {NXTLightDetector . NXTRangeDetector . NXTTouchSensor . NXTSoundSensor . NXTCompassSensor . NXTGyroSensor}
	].
	(dialogName = 'motors') ifTrue: [
		^ {NXTMotor . NXTDCMotor}
	].
	(dialogName = 'driving') ifTrue: [
		^ {NXTDifferentialPilot}
	].

	^ #()! !


!NXTConfiguration class methodsFor: 'block specs' stamp: 'CWB 12/21/2012 19:45'!
deviceBlockSpecsFor: blockSpecRequester
"Override this to supply block specs that should always be active for this device."

"blockSpecRequester can be #stage, #sprite, or #other"

	| specs inapplicableSpecs common1 sprite2 stage2 common3 sprite4 stage4 |

	common1 _ #(
		'sound'
			"Commented out lines are pending development, Remove them from final if still commented out."
			('play sound %S'				-	playSound:)
			('play sound %S until done'		s	doPlaySoundAndWait)
			"-"
			('rest for %n beats'				t	rest:elapsed:from: 0.5)
			-
			('play note %N for %n beats'	t	noteOn:duration:elapsed:from: 60 0.5)
			('set instrument to %I'			- 	nxtMidiInstrument: 1)
			('instrument'					r	nxtMidiInstrument)
			-
			('change volume by %n'		- 	nxtChangeVolumeBy: -10)
			('set volume to %n%'			- 	nxtSetVolumeTo: 100)
			('volume'						r 	nxtVolume)
			-
			('change tempo by %n'			- 	nxtChangeTempoBy: 20)
			('set tempo to %n bpm'			-	nxtSetTempoTo: 60)
			('tempo'							r 	nxtTempo)
	).

	sprite2 _ #(
		'looks'
			('switch to costume %l'			-	lookLike:)
			('next costume'					-	nextCostume)
			('costume #'						r	costumeIndex)
			-
			('say %s for %n secs'			t	say:duration:elapsed:from: 'Hello!!' 2)
			('say %s'						-	say: 'Hello!!')
			('print %s'						-	nxtPrint: 'Hello')
			-
			('show'							-	show)
			('hide'							-	hide)
	).

	stage2 _ #(
		'looks'
			('switch to background %l'		-	showBackground: 'background1')
			('next background'				-	nextBackground)
			('background #'					r	backgroundIndex)
			-
			('print %s'						-	nxtPrint: 'Hello')
	).

	common3 _ #(
		
		'sensing'	
			('battery voltage'				r	nxtBatteryVoltage)
			-
			('%H sensor value'			r	sensor: '(remote)')
			"('sensor %h?'				b	sensorPressed: 'button pressed')"
			-
			('reset timer'					-	timerReset)
			('timer'							r	timer)
			-
			('%a of %m'					r	getAttribute:of:)
			-
			('touching %m?' 				b	touching:)
			('touching color %C?' 			b	touchingColor:)
			('color %C is touching %C?'		b	color:sees:)

 		'motion'		"stage doesn't show any motion blocks, so we can include these."
 			('move %n steps'				-	forward:)
			('turn %J %n degrees'			-	NXTPilotX2:onScreenTurnDegrees: 'img:turnCW' 15)	
 			-
 			('point in direction %d'			-	heading: 90)
 			('point towards %m'				-	pointTowards:)
 			-
 			('go to x:%n y:%n'				-	gotoX:y: 0 0)
 			('go to %m'						-	gotoSpriteOrMouse:)
 			('glide %n secs to x:%n y:%n'	t	glideSecs:toX:y:elapsed:from: 1 50 50)
 			-
 			('change x by %n'				-	changeXposBy: 10)
 			('set x to %n'					-	xpos: 0)
 			('change y by %n'				-	changeYposBy: 10)
 			('set y to %n'					-	ypos: 0)
 			-
 			('if on edge, bounce'			-	bounceOffEdge)
 			-
 			('x position'						r	xpos)
 			('y position'						r	ypos)
 			('direction'						r	heading)

		'operators'
				('map %n <br> from: [%n, %n] <br> to: [%n, %n]'		r	mapValue:fromLow:fromHigh:toLow:toHigh: 180 0 360 0 100)
				-


		'pen'
			('clear'							-	clearPenTrails)

	).
	
	sprite4 _ #(
			-
			('pen down'						-	putPenDown)
			('pen up'						-	putPenUp)
			-
			('set pen color to %c'			-	penColor:)
			-
			('stamp'							-	stampCostume)
	).
	stage4 _ #().

	inapplicableSpecs _ #(
		('when %k key pressed' #K #-)
		('when %m clicked' #M #-)
		('broadcast %e and wait' #s #doBroadcastAndWait)
	).


	(blockSpecRequester = #sprite)	ifTrue: [ specs _ common1, sprite2, common3, sprite4 ].
	(blockSpecRequester = #stage)	ifTrue: [ specs _ common1, stage2, common3, stage4 ].
	(blockSpecRequester = #all)		ifTrue: [ specs _ common1, sprite2, stage2, common3, sprite4, stage4 ].


	"return all specs that aren't inapplicable"
	^ specs, self standardBlockSpecs 
		reject: [:spec | inapplicableSpecs includes: spec ]. 


! !
