NXTPilot subclass: #NXTDifferentialPilot
	instanceVariableNames: 'wheelDiameter wheelDiameterUnits trackWidth trackWidthUnits leftMotorPort rightMotorPort reverseMotors '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Enchanting-NXT Peripherals'!

!NXTDifferentialPilot methodsFor: 'conversion' stamp: 'CWB 11/23/2012 18:08'!
createConfigBlock
"A peripheral should create a config block and fill it in with pertinent information"

	| block |

	block _ self class createGenericConfigBlock.

	(block argumentAt: 1) numExpression: wheelDiameter.
	(block argumentAt: 2) choice: wheelDiameterUnits.
	(block argumentAt: 3) numExpression: trackWidth.
	(block argumentAt: 4) choice: trackWidthUnits.
	(block argumentAt: 5) choice: leftMotorPort.
	(block argumentAt: 6) choice: rightMotorPort.
	(block argumentAt: 7) choice: (self reversedAsText).

	^ block.

! !

!NXTDifferentialPilot methodsFor: 'conversion' stamp: 'CWB 11/23/2012 18:08'!
takeSettingsFrom: configBlock
"With a morph of the appropriate class, this gets the necessary settings from the configuration block"

	wheelDiameter _ (configBlock argumentAt: 1) evaluate.
	wheelDiameterUnits _ (configBlock argumentAt: 2) choice.
	trackWidth _ (configBlock argumentAt: 3) evaluate.
	trackWidthUnits _ (configBlock argumentAt: 4) choice.
	leftMotorPort _ (configBlock argumentAt: 5) choice.
	rightMotorPort _ (configBlock argumentAt: 6) choice.
	self reversedFrom: (configBlock argumentAt: 7) choice.

	^ self.

! !


!NXTDifferentialPilot methodsFor: 'initialization' stamp: 'CWB 11/23/2012 18:10'!
initialize

	| defaults |

	super initialize.

	defaults _ self class defaultArgs.
	
	wheelDiameter _ defaults at: 1.
	wheelDiameterUnits _ defaults at: 2.
	trackWidth _ defaults at: 3.
	trackWidthUnits _ defaults at: 4.
	leftMotorPort _ defaults at: 5.
	rightMotorPort _ defaults at: 6.
	self reversedFrom: (defaults at: 7).

	! !


!NXTDifferentialPilot methodsFor: 'object i/o' stamp: 'CWB 11/17/2012 21:04'!
fieldsVersion

	^ 1
! !

!NXTDifferentialPilot methodsFor: 'object i/o' stamp: 'CWB 11/23/2012 18:14'!
initFieldsFrom: anObjStream version: classVersion

	"super initFieldsFrom: anObjStream version: classVersion."
	self initFieldsNamed: #(
		wheelDiameter
		wheelDiameterUnits 
		trackWidth
		trackWidthUnits
		leftMotorPort 
		rightMotorPort
		reverseMotors
	) from: anObjStream.! !

!NXTDifferentialPilot methodsFor: 'object i/o' stamp: 'CWB 11/23/2012 18:14'!
storeFieldsOn: anObjStream

	"super storeFieldsOn: anObjStream."
	self storeFieldsNamed: #(
		wheelDiameter
		wheelDiameterUnits 
		trackWidth
		trackWidthUnits
		leftMotorPort 
		rightMotorPort
		reverseMotors
	) on: anObjStream.
! !


!NXTDifferentialPilot methodsFor: 'exporting' stamp: 'CWB 11/23/2012 19:16'!
emitInitializationCodeOn: aStream for: anExporter
"Outputs the code to initialize NXT peripherals."

"Output should look something like:
	
	public static final DifferentialPilot pilot = new DifferentialPilot(43.2f, 118.0f, Motor.A, Motor.B, false);
	public static final Navigator navigator = new Navigator(pilot);	

	Note that all sizes will be expressed in millimeters.
"

	| wheelDiameterInMM trackWidthInMM  |

	wheelDiameterInMM _ self length: wheelDiameter toMMfrom: wheelDiameterUnits.
	trackWidthInMM _ self length: trackWidth toMMfrom: trackWidthUnits.

	anExporter indentOn: aStream.
	aStream nextPutAll: 'public static final DifferentialPilot pilot = new DifferentialPilot('.
	anExporter exportLiteralValue: wheelDiameterInMM on: aStream.
	aStream nextPutAll: ', '.
	anExporter exportLiteralValue: trackWidthInMM on: aStream.
	aStream nextPutAll: ', ', (self encodedMotorNameFor: leftMotorPort).
	aStream nextPutAll: ', ', (self encodedMotorNameFor: rightMotorPort), ', '.
	anExporter exportLiteralValue: reverseMotors on: aStream.
	aStream nextPutAll: ');'; cr.

	anExporter indentOn: aStream.
	aStream nextPutAll: 'public static final Navigator navigator = new Navigator(pilot);'; cr.

	
	
		! !


!NXTDifferentialPilot methodsFor: 'private' stamp: 'CWB 11/19/2012 11:59'!
encodedMotorNameFor: motorPort
"Takes a motor port, like 'Port A', and returns 'Motor.A', for use in code"

	^ 'Motor.', motorPort last asString.! !

!NXTDifferentialPilot methodsFor: 'private' stamp: 'CWB 11/19/2012 11:50'!
length: length toMMfrom: units
"Takes a length and some units and converts the results to millimeters"

	(units = 'mm') ifTrue: [ ^ length ].
	(units = 'cm') ifTrue: [ ^ length * 10 ].
	(units = 'in') ifTrue: [ ^ length * 25.4 ].
	^ self error: 'Unknown units'
	
! !

!NXTDifferentialPilot methodsFor: 'private' stamp: 'CWB 11/23/2012 18:07'!
reversedAsText

	reverseMotors 
		ifTrue: [ ^ 'in reverse' ] 		
		ifFalse: [ ^ 'forwards' ].! !

!NXTDifferentialPilot methodsFor: 'private' stamp: 'CWB 11/23/2012 18:07'!
reversedFrom: aString

	reverseMotors _ (aString = 'in reverse').
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

NXTDifferentialPilot class
	instanceVariableNames: ''!

!NXTDifferentialPilot class methodsFor: 'accessing' stamp: 'CWB 9/15/2012 22:16'!
baseName

	^ 'differential pilot'! !

!NXTDifferentialPilot class methodsFor: 'accessing' stamp: 'CWB 9/15/2012 22:16'!
concreteTypes
"Returns a list of physical devices that implement the functions of this class."


	^ #(
		'Differential Pilot'
	).! !


!NXTDifferentialPilot class methodsFor: 'block specs' stamp: 'CWB 11/23/2012 17:44'!
configBlockSpec

	^ 'Differential drive <br> wheel diameter: %n %J <br> track width: %n %J <br> left motor: %K <br> right motor: %K <br> motors drive %Q'.! !

!NXTDifferentialPilot class methodsFor: 'block specs' stamp: 'CWB 11/23/2012 18:11'!
defaultArgs

	"Return the default arguments."

	^ { 43.2 . 'mm' . 17.6 . 'cm' . 'Port B' . 'Port C' . 'forwards' }.
! !

!NXTDifferentialPilot class methodsFor: 'block specs' stamp: 'CWB 12/20/2012 11:20'!
providesBlocks
"Returns a list of block specifications to be used in the main user interface if there is any device configured that can use these blocks."

	| blocks |
	
	blocks _ #(
		'driving'
			('drive %K'							-	NXTPilotX1:drive:)
			('drive %K %n %J'					-	NXTPilotX1:drive:distance: '-' 10 'cm')
			('drive %K steering: %q'				-	NXTPilotX1:steer: '-' 100)
			-
			('stop'								-	NXTPilotX1:stop:)
			-
			('turn %J'							-	NXTPilotX2:rotate:)
			('turn %J %n'						-	NXTPilotX2:rotate:degrees: '-' 15)
			('point in direction %d'				-	NXTPilotX1:pointInDirection: 90)
			-
			('arc %Q radius: %n %J'				-	NXTPilotX1:arc: '-' 50 'cm')
			('arc %Q radius: %n %J for %n'		-	NXTPilotX1:arc:degrees: '-' 50 'cm' 360)
			-
			('set driving speed to %n %J/s'		-	NXTPilotX1:setDrivingSpeed: 30 'cm')
			('set acceleration to %n %J/s/s'		-	NXTPilotX1:setAcceleration: 120 'cm')
			('set turning speed to %n/s'			-	NXTPilotX1:setTurningSpeed: 300)
			-
			('driving speed in %J/s'				r	NXTPilotX1:drivingSpeed: 'cm')
			('turning speed in /s'				r	NXTPilotX1:turningSpeed:)
			~
			('fastest driving speed in %J/s'		r	NXTPilotX1:fasterDrivingSpeed: 'cm')
			('fastest turning speed in /s'		r	NXTPilotX1:fastestTurningSpeed:)
			-
			('go to x: %n %J y: %n %J'			-	NXTPilotX1:gotoX:y: 0 'cm' 0 'cm')
			('go to <br> x: %n %J <br> y: %n %J <br> at %d'	-	NXTPilotX1:gotoX:y:heading: 0 'cm' 0 'cm' 90)
			-
			('is moving?'						b	NXTPilotX1:isMoving:)
			('x position in %J'					r	NXTPilotX1:xpos: 'cm')
			('y position in %J'					r	NXTPilotX1:ypos: 'cm')
			('direction in '						r	NXTPilotX1:heading:)
			-
			('pose is now <br> x: %n %J <br> y: %n %J <br> at %d'	-	NXTPilotX1:changePoseTo: 0 'cm' 0 'cm' 90)
			('x position is now %n %J'			-	NXTPilotX1:changeXposTo: 0 'cm')
			('y position is now %n %J'			-	NXTPilotX1:changeYposTo: 0 'cm')
			('direction is now %d'				-	NXTPilotX1:changeHeading: 90)
		).

	^ blocks  ! !


!NXTDifferentialPilot class methodsFor: 'dropdown lists' stamp: 'CWB 11/23/2012 17:45'!
customListNumber: aNumber
"Peripherals can return custom lists (or nil)."

	(aNumber = 1) ifTrue: [
		"units of measure"
		^ #('mm' 'cm' 'in')	
	].
	(aNumber = 2) ifTrue: [
		"motor port"
		^ #( 'Port A' 'Port B' 'Port C' ).
	].
	(aNumber = 3) ifTrue: [
		"motor direction"
		^ #( 'forwards' 'in reverse' ).
	].

	^ nil.! !
